/*
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author: Adrian Szyndela <adrian.s@samsung.com>
 */
/**
 * @file crash-stack-libelf.c
 * @brief unwinding call stacks, functions specific for archs that use only libelf
 */
#include "crash-stack.h"
#include <elfutils/libdwfl.h>
#include <elfutils/version.h>
#include <string.h>

/**
 * @brief Convenience type for registers of different sizes
 */
typedef union {
	uint16_t reg16;		///< 16-bit register
	uint32_t reg32;		///< 32-bit register
	uint64_t reg64;		///< 64-bit register
} Register;

static Register g_pc;	///< storage for program counter value
static Register g_sp;	///< storage for stack pointer value

#if _ELFUTILS_PREREQ(0,158)
/**
 * @brief Callback for modern elfutils; called on each found frame.
 *
 * @param state state of the calls tack traversal
 * @param arg user data - pointer to callstack database
 */
static int __frame_callback(Dwfl_Frame *state, void *arg)
{
	Callstack *callstack = (Callstack*)arg;
	Dwarf_Addr address;
	dwfl_frame_pc(state, &address, NULL);
	callstack->tab[callstack->elems++] = address;
	return callstack->elems < MAX_CALLSTACK_LEN ? DWARF_CB_OK : DWARF_CB_ABORT;
}

/**
 * @brief Callback for modern elfutils; called on each found thread.
 *
 * @param thread thread handle
 * @param thread_arg user data - pointer to callstack database
 */
static int __thread_callback(Dwfl_Thread *thread, void *thread_arg)
{
	dwfl_thread_getframes(thread, __frame_callback, thread_arg);
	return DWARF_CB_ABORT;
}
#endif

/**
 * @brief common names for program counter/instruction pointer
 */
static const char *pc_names[] = {
  "pc", "rip", "eip", "ip"
};

/**
 * @brief common names for stack pointer
 */
static const char *sp_names[] = {
  "sp", "rsp", "esp"
};

/**
 * @brief Helper functions that checks if a name is present in array of names
 *
 * @param name name to look for
 * @param names array of names
 * @param elems number of elements in names
 * @returns true if name is found in names, false otherwise
 */
static bool __is_in(const char *name, const char **names, int elems)
{
	int nit;
	for (nit = 0; nit < elems; ++nit) {
		if (strcmp(name, names[nit]) == 0)
			return true;
	}
	return false;
}

/**
 * @brief Helper macro for looking for name in array of names
 */
#define IS_IN(name,names) __is_in((name), (names), sizeof(names)/sizeof(names[0]))

void *_get_place_for_register_value(const char *regname, int regnum)
{
	if (IS_IN(regname, pc_names)) return &g_pc;
	else if (IS_IN(regname, sp_names)) return &g_sp;

	return 0;
}

/**
 * @brief Tries to heuristically gather call stack information.
 *
 * @remarks This function walks through the stack of the given process, and checks
 *          if found addresses are addresses belonging to functions. If so, it treats
 *          such address as an element of the call stack.
 *          While it gathers some useful information, it gathers also "false positives".
 * @remarks It should be used as a last resort - when no other method is available.
 *
 * @param dwfl dwfl handle
 * @param core core file handle, NULL if live process is analyzed
 * @param pid PID of the analyzed process, 0 if core dump file is analyzed
 * @param mappings module mappings database
 * @param callstack call stack database
 */
static void _explore_stack_in_search_of_functions(Dwfl *dwfl, Elf *core, pid_t pid,
				Mappings *mappings, Callstack *callstack)
{
	Dwarf_Addr stack_pointer = sizeof(uintptr_t) == 4 ? g_sp.reg32 : g_sp.reg64;
	Dwarf_Addr stack_max_lookup = stack_pointer + 8*1024*1024;
	bool data_remaining = true;
	do {
		uintptr_t value;
		data_remaining = _crash_stack_libelf_read_value(dwfl, core, pid, stack_pointer,
														&value, sizeof(value), mappings);
		if (data_remaining) {
			Dwarf_Addr bias;
			/* check presence of address in text sections */
			Dwfl_Module *module = dwfl_addrmodule(dwfl, value);
			if (module != NULL) {
				Dwarf_Addr elfval = value;
				GElf_Sym sym;
				GElf_Word shndxp = -1;
				dwfl_module_addrsym(module, value, &sym, &shndxp);
				Elf_Scn *scn = dwfl_module_address_section(module, &elfval, &bias);
				if (scn != 0 || shndxp != -1) {
					callstack->tab[callstack->elems++] = value;
					if (callstack->elems >= MAX_CALLSTACK_LEN)
						return;
				}
				else {
					/* check also inside [pie] and [exe] */
					int i;
					for (i = 0; i < mappings->elems; i++) {
						if (mappings->tab[i].m_start <= elfval && elfval < mappings->tab[i].m_end) {
							callstack->tab[callstack->elems++] = value;
							break;
						}
					}
				}
			}
		}
		stack_pointer += sizeof(uintptr_t);
	} while (data_remaining && stack_pointer < stack_max_lookup);
}

void _create_crash_stack(Dwfl *dwfl, Elf *core, pid_t pid, Mappings *mappings, Callstack *callstack)
{
	callstack->elems = 0;
#if _ELFUTILS_PREREQ(0,158)
	dwfl_getthreads(dwfl, __thread_callback, callstack);
#else
	callstack->tab[callstack->elems++] = sizeof(uintptr_t) == 4 ? g_pc.reg32 : g_pc.reg64;
	_explore_stack_in_search_of_functions(dwfl, core, pid, mappings, callstack);
#endif
}

