/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack-x86_64.c
 * @brief unwinding call stacks, functions specific for x86_64
 */
#include "crash-stack.h"
#include <sys/user.h>
#include <string.h>

struct user_regs_struct g_registers;	///< static storage for ptrace buffer for registers

void *_crash_stack_get_memory_for_ptrace_registers(size_t *size)
{
	if (NULL != size)
		*size = sizeof(g_registers);
	return &g_registers;
}

void _crash_stack_set_ptrace_registers(void *regbuf)
{
	void *rsp = _get_place_for_register_value("rsp", 0);
	void *rip = _get_place_for_register_value("rip", 0);

	struct user_regs_struct *regs = regbuf;

	memcpy(rsp, &regs->rsp, sizeof(regs->rsp));
	memcpy(rip, &regs->rip, sizeof(regs->rip));
}

void _crash_stack_print_regs(FILE* outputfile)
{
#define _PRINT_REGISTERS(a,b,c)						\
	fprintf(outputfile, "%3s: %016llx %3s: %016llx %3s: %016llx\n", \
		#a, g_registers.a, \
		#b, g_registers.b, \
		#c, g_registers.c)
	fprintf(outputfile, "\nRegister Information\n");
#ifdef __x86_64__
	fprintf(outputfile, "rip: %04llx:[<%016llx>]\n",
		g_registers.cs & 0xffff,
		g_registers.rip);
	fprintf(outputfile, "rsp: %04llx:%016llx eflags: %08llx\n",
		g_registers.ss,
		g_registers.rsp,
		g_registers.eflags);
	_PRINT_REGISTERS(rax, rbx, rcx);
	_PRINT_REGISTERS(rdx, rsi, rdi);
	_PRINT_REGISTERS(rbp, r8, r9);
	_PRINT_REGISTERS(r10, r11, r12);
	_PRINT_REGISTERS(r13, r14, r15);
#else
	fprintf(outputfile, "Unsupported architecture\n")
#endif
	fprintf(outputfile, "\n");
#undef _PRINT_REGISTERS
}
