/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/**
 * @file crash-stack.c
 * @brief This file contains Main module of call stack unwinding program
 *
 * Crash-stack is a single purpose program. Its duty is to show call stack
 * of a crashed program. Crash-stack must be called with proper arguments:
 * either core dump file name or PID of a crashed program.
 */
#include <stdlib.h>
#include <stdio.h>
#include <libelf.h>
#include <elfutils/libdwfl.h>
/*
 * In case the program is compiled with core dumps, the license may switch to GPL2.
 */
#ifdef WITH_CORE_DUMP
#include <elfutils/libebl.h>
#endif
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/prctl.h>
#include <linux/prctl.h>
#include "crash-stack.h"
#include <string.h>
#include <elfutils/version.h>
#include <getopt.h>
#include <sys/ptrace.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/wait.h>

static siginfo_t __siginfo;
static FILE *outputfile = NULL;		///< global output stream
static FILE *errfile = NULL;		///< global error stream

/**
 * @brief definitions for getopt options: identifiers
 */
enum {
	OPT_PID,
	OPT_OUTPUTFILE,
	OPT_ERRFILE
};

/**
 * @brief definitions for getopt options: full specifications
 */
const struct option opts[] = {
	{ "pid", required_argument, 0, OPT_PID },
	{ "output", required_argument, 0, OPT_OUTPUTFILE },
	{ "erroutput", required_argument, 0, OPT_ERRFILE },
	{ 0, 0, 0, 0 }
};

/*
 * __cxa_demangle() is taken from libstdc++, however there is no header that we
 * can take a declaration from. Importing through 'extern' allows using it.
 */
/// @cond false
extern char *__cxa_demangle(const char *mangled_name, char *output_buffer,
		size_t *length, int *status);
///@endcond

/**
 * @brief A callback for dwfl_getmodules().
 *
 * This callback is called once for every module discovered by dwfl_getmodules().
 *
 * @param module the dwfl module
 * @param userdata unused, required by dwfl_getmodules()
 * @param name name of the module
 * @param address address of the module
 * @param arg 4th argument to dwfl_getmodules is passed here
 */
static int __module_callback(Dwfl_Module *module, void **userdata,
		const char *name, Dwarf_Addr address,
		void *arg)
{
	if (name != NULL && name[0] == '[') {
		/* libdwfl couldn't get the module file - we will get it later from notes */
		Mappings *mappings = arg;
		if (mappings->elems < MAX_MAPPINGS_NUM) {
			size_t elems = mappings->elems;
			mappings->tab[elems].m_start = address;
			mappings->tab[elems].m_end = 0;
			mappings->tab[elems].m_offset = 0;
			mappings->tab[elems].m_name = NULL;
			mappings->tab[elems].m_fd = -1;
			mappings->tab[elems].m_elf = 0;
			mappings->elems++;
		}
	}
	return DWARF_CB_OK;
}

/**
 * @brief Reads a value of specified size from core dump file.
 *
 * @param core ELF handler for the core dump file
 * @param from the source address in the ELF file
 * @param size size of the value in bits
 * @param to the address of allocated memory, where the value will be copied
 */
static void __get_value(Elf *core, const void *from, size_t size, void *to)
{
	Elf_Type type = ELF_T_BYTE;
	switch (size) {
	case 8: type = ELF_T_BYTE; break;
	case 16: type = ELF_T_HALF; break;
	case 32: type = ELF_T_WORD; break;
	case 64: type = ELF_T_XWORD; break;
	default:
		 fprintf(stderr, "__get_value for strange size: %llu\n", (unsigned long long)size);
		 break;
	}
	Elf_Data out = {
		.d_buf = to,
		.d_type = type,
		.d_version = EV_CURRENT,
		.d_size = size/8,
		.d_off = 0,
		.d_align = 0
	};
	Elf_Data in = {
		.d_buf = (void*)(from),
		.d_type = out.d_type,
		.d_version = out.d_version,
		.d_size = out.d_size,
		.d_off = 0,
		.d_align = 0
	};
	Elf_Data *data;
	if (gelf_getclass(core) == ELFCLASS32)
		data = elf32_xlatetom(&out, &in, elf_getident(core, NULL)[EI_DATA]);
	else
		data = elf64_xlatetom(&out, &in, elf_getident(core, NULL)[EI_DATA]);
	if (data == NULL)
		fprintf(errfile, "failed to get value from core file\n");
}

/**
 * @brief gets number of values, page size, address size, values and names from ELF notes
 *
 * @remarks This is very specific for organization of notes part in ELF files
 *
 * @param elf ELF handler - may be core file
 * @param desc address of ELF notes descriptor
 * @param[out] values_cnt a place for number of values
 * @param[out] page_size a place for page size
 * @param[out] addr_size a place for address size
 * @param[out] values a place for address of values
 * @param[out] filenames a place for address of filenames
 */
static void __parse_note_file(Elf *elf, const char *desc, uint64_t *values_cnt, uint64_t *page_size,
		size_t *addr_size, const char **values, const char **filenames)
{
	*addr_size = gelf_fsize(elf, ELF_T_ADDR, 1, EV_CURRENT);
	__get_value(elf, desc, *addr_size*8, values_cnt);
	__get_value(elf, desc + *addr_size, *addr_size*8, page_size);
	/* First: triplets of <mapping-start> <mapping-end> <offset-in-pages>
	 *     count = values_cnt
	 * Then the names of files.
	 */
	*values = desc + 2 * *addr_size;
	*filenames = *values + 3 * *addr_size * *values_cnt;
}

/**
 * @brief Simple accessor for mapping items in ELF files
 *
 * @remarks This is very specific for organization of notes part in ELF files
 *
 * @param elf ELF handler - may be core file
 * @param addr_size size of addresses in this ELF
 * @param item address of mapping item to get data from
 * @param[out] mapping_start value of the start of the mapping
 * @param[out] mapping_end value of the end of the mapping
 * @param[out] offset_in_pages number of pages of offset in the file
 */
static void __get_mapping_item(Elf *elf, size_t addr_size, const void *item,
		uint64_t *mapping_start, uint64_t *mapping_end, uint64_t *offset_in_pages)
{
	__get_value(elf, item, addr_size*8, mapping_start);
	__get_value(elf, item + addr_size, addr_size*8, mapping_end);
	__get_value(elf, item + 2 * addr_size, addr_size*8, offset_in_pages);
}

/**
 * @brief Tries to get symbol name from ELF files
 *
 * @remarks This function is used in case that symbol name is not available by libelf,
 *			e.g. when old libelf version does not take into account some modules.
 *
 * @param core ELF handler for the core dump file
 * @param notes notes descriptor from core file
 * @param address address to get symbol name for
 * @param[out] module_name name of the module of the found symbol. Not touched when
 *             symbol is not found
 * @return name of the symbol or NULL if not found
 */
static char *__try_symbol_from_elfs(Elf *core, Elf_Data *notes, uintptr_t address,
		const char **module_name)
{
	GElf_Nhdr nhdr;
	char *symbol = NULL;
	size_t pos = 0;
	size_t new_pos = 0;
	size_t name_pos;
	size_t desc_pos;

	while ((new_pos = gelf_getnote(notes, pos, &nhdr, &name_pos, &desc_pos)) > 0) {
		if (nhdr.n_type == NT_FILE) {
			uint64_t values_cnt = 0, page_size = 0;
			const char *values;
			const char *filenames;
			size_t addr_size = 0;

			__parse_note_file(core, notes->d_buf + desc_pos, &values_cnt,
							  &page_size, &addr_size, &values, &filenames);

			int ii;
			for (ii = 0; ii < values_cnt; ii++) {
				uint64_t mapping_start = 0, mapping_end = 0, offset_in_pages = 0;
				const char *item = values + 3 * addr_size * ii;

				__get_mapping_item(core, addr_size, item, &mapping_start, &mapping_end,
								 &offset_in_pages);

				if (mapping_start <= address && address < mapping_end) {
					Elf *elf;
					int fd;
					fd = open(filenames, O_RDONLY);
					if (-1 == fd)
						return NULL;

					elf = elf_begin(fd, ELF_C_READ_MMAP, NULL);

					if (NULL == elf) {
						close(fd);
						return NULL;
					}

					Elf_Scn *scn = NULL;
					*module_name = filenames;

					while ((scn = elf_nextscn(elf, scn)) != NULL) {
						GElf_Shdr shdr_mem;
						GElf_Shdr *shdr = gelf_getshdr(scn, &shdr_mem);
						if (shdr != NULL && (shdr->sh_type == SHT_SYMTAB || shdr->sh_type == SHT_DYNSYM)) {
							Elf_Data *sdata = elf_getdata(scn, NULL);
							unsigned int nsyms = sdata->d_size / (gelf_getclass(elf) == ELFCLASS32 ?
									sizeof(Elf32_Sym) :
									sizeof(Elf64_Sym));
							unsigned int cnt;
							uintptr_t address_offset = address;
							if (shdr->sh_type == SHT_DYNSYM)
								address_offset -= mapping_start;
							for (cnt = 0; cnt < nsyms; ++cnt) {
								GElf_Sym sym_mem;
								Elf32_Word xndx;
								GElf_Sym *sym = gelf_getsymshndx(sdata, NULL, cnt, &sym_mem, &xndx);
								if (sym != NULL && sym->st_shndx != SHN_UNDEF) {
									if (sym->st_value <= address_offset && address_offset < sym->st_value + sym->st_size) {
										symbol = strdup(elf_strptr(elf, shdr->sh_link, sym->st_name));
										break;
									}
								}
							}
						}
					}

					elf_end(elf);
					close(fd);
					return symbol;
				}

				filenames += strlen(filenames)+1;
			}
		}
		pos = new_pos;
	}

	return NULL;
}

/**
 * @brief Opens libdwfl for using with live process
 *
 * @param pid pid of the process to attach to
 * @return Dwfl handle
 */
static Dwfl *__open_dwfl_with_pid(pid_t pid)
{
	int status;
	pid_t stopped_pid;

	if (ptrace(PTRACE_SEIZE, pid, NULL, PTRACE_O_TRACEEXIT) != 0) {
		fprintf(errfile, "PTRACE_SEIZE failed on PID %d: %m\n", pid);
		return NULL;
	}


	ptrace(PTRACE_INTERRUPT, pid, 0, 0);

	stopped_pid = waitpid(pid, &status, 0);
	if (stopped_pid == -1 || stopped_pid != pid || !WIFSTOPPED(status)) {
		fprintf(errfile, "waitpid failed: %m, stopped_pid=%d, status=%d\n", stopped_pid, status);
		return NULL;
	}

	if (ptrace(PTRACE_GETSIGINFO, pid, NULL, &__siginfo) != 0)
		return NULL;

	static const Dwfl_Callbacks proc_callbacks = {
		.find_elf = dwfl_linux_proc_find_elf,
		.find_debuginfo = dwfl_standard_find_debuginfo,
		.section_address = NULL,
		.debuginfo_path = NULL
	};

	Dwfl *dwfl = dwfl_begin(&proc_callbacks);
	if (dwfl == NULL) {
		fprintf(errfile, "process %d : Can't start dwfl (%s)\n", pid, dwfl_errmsg(-1));
		return NULL;
	}

	if (dwfl_linux_proc_report(dwfl, pid) < 0) {
		fprintf(errfile, "process %d : dwfl report failed (%s)\n", pid, dwfl_errmsg(-1));
		dwfl_end(dwfl);
		return NULL;
	}

#if _ELFUTILS_PREREQ(0,158)
	if (dwfl_linux_proc_attach(dwfl, pid, true) < 0) {
		fprintf(errfile, "process %d : dwfl attach failed (%s)\n", pid, dwfl_errmsg(-1));
		dwfl_end(dwfl);
		return NULL;
	}
#endif
	return dwfl;
}

/**
 * @brief Opens libdwfl for using with core dump file
 *
 * @remarks This function will open core file regardless of WITH_CORE_DUMP setting.
 *          It may help with detecting issues with using dwfl even without support
 *          for dumps.
 *
 * @param core elf handler for the core dump file
 * @param core_file_name name of the core file; needed only for diagnostics
 * @return Dwfl handle
 */
static Dwfl *__open_dwfl_with_core(Elf *core, const char *core_file_name)
{
	static const Dwfl_Callbacks core_callbacks = {
		.find_elf = dwfl_build_id_find_elf,
		.find_debuginfo = dwfl_standard_find_debuginfo,
		.section_address = NULL,
		.debuginfo_path = NULL
	};

	Dwfl *dwfl = dwfl_begin(&core_callbacks);
	if (dwfl == NULL) {
		fprintf(errfile, "%s : Can't start dwfl (%s)\n", core_file_name, dwfl_errmsg(-1));
		return NULL;
	}

#if _ELFUTILS_PREREQ(0,158)
	if (dwfl_core_file_report(dwfl, core, NULL) < 0)
#else
		if (dwfl_core_file_report(dwfl, core) < 0)
#endif
		{
			fprintf(errfile, "%s : dwfl report failed (%s)\n", core_file_name, dwfl_errmsg(-1));
			dwfl_end(dwfl);
			return NULL;
		}

#if _ELFUTILS_PREREQ(0,158)
	if (dwfl_core_file_attach(dwfl, core) < 0) {
		fprintf(errfile, "%s : dwfl attach failed (%s)\n", core_file_name, dwfl_errmsg(-1));
		dwfl_end(dwfl);
		return NULL;
	}
#endif
	return dwfl;
}

/**
 * @brief Gets registers information for live process
 *
 * @param pid pid of the live process
 * @return 0 on success, -1 otherwise
 */
static int __get_registers_ptrace(pid_t pid)
{
	struct iovec data;
	data.iov_base = _crash_stack_get_memory_for_ptrace_registers(&data.iov_len);

	if (NULL == data.iov_base) {
		fprintf(errfile, "Cannot get memory for registers for ptrace (not implemented for this architecture\n");
		return -1;
	}

	if (ptrace(PTRACE_GETREGSET, pid, NT_PRSTATUS, &data) != 0) {
		fprintf(errfile, "PTRACE_GETREGSET failed on PID %d: %m\n", pid);
		return -1;
	}

	_crash_stack_set_ptrace_registers(data.iov_base);

	return 0;
}

/**
 * @brief Get information about the signal that stopped the process
 *
 * @param pid pid of the live process
 * @return 0 on success, -1 otherwise
 */
static int __get_signal_ptrace(pid_t pid)
{
	const char* const signal_table[] = {
		[SIGHUP]="SIGHUP", [SIGINT]="SIGINT", [SIGQUIT]="SIGQUIT",
		[SIGILL]="SIGILL", [SIGTRAP]="SIGTRAP", [SIGABRT]="SIGABRT",
		[SIGIOT]="SIGIOT", [SIGBUS]="SIGBUS", [SIGFPE]="SIGFPE",
		[SIGKILL]="SIGKILL", [SIGUSR1]="SIGUSR1", [SIGSEGV]="SIGSEGV",
		[SIGUSR2]="SIGUSR2", [SIGPIPE]="SIGPIPE", [SIGALRM]="SIGALRM",
		[SIGTERM]="SIGTERM", [SIGSTKFLT]="SIGSTKFLT", [SIGCHLD]="SIGCHLD",
		[SIGCONT]="SIGCONT", [SIGSTOP]="SIGSTOP", [SIGTSTP]="SIGTSTP",
		[SIGTTIN]="SIGTTIN", [SIGTTOU]="SIGTTOU", [SIGURG]="SIGURG",
		[SIGXCPU]="SIGXCPU", [SIGXFSZ]="SIGXFSZ", [SIGVTALRM]="SIGVTALRM",
		[SIGPROF]="SIGPROF", [SIGWINCH]="SIGWINCH", [SIGIO]="SIGIO",
		[SIGPWR]="SIGPWR", [SIGSYS]="SIGSYS", [SIGUNUSED]="SIGUNUSED",
	};

	printf("Signal: %d\n"
	       "\t(%s)\n"
	       "\tsi_code: %d\n",
	       __siginfo.si_signo,
	       signal_table[__siginfo.si_signo],
	       __siginfo.si_code);
	switch (__siginfo.si_code) {
	case SI_TKILL:
	case SI_USER:
		printf("\tsignal sent by %s (sent by pid %d, uid %d)",
		       __siginfo.si_code == SI_TKILL ? "tkill" : "kill",
		       __siginfo.si_pid, __siginfo.si_uid);
		break;
	case SI_KERNEL:
		printf("\tsignal sent by the kernel\n");
		break;
	}
	return 0;
}

#ifdef WITH_CORE_DUMP
/**
 * @brief Helper function for updating mappings.
 *
 * @remarks Old versions of libelf not always extract full information about modules.
 *          For such cases we maintain mappings for every module. Those mappings
 *          may be updated while reading notes from core file.
 *
 * @param mappings mappings database
 * @param mapping_start address of the mapped start of the module; module is identified
 *                      by mapping_start
 * @param mapping_end address of the end of the module; needed to compute module boundaries
 * @param offset offset within core file - unused
 * @param name file name of the module
 */
static void __updateMapping(Mappings *mappings, uint64_t mapping_start, uint64_t mapping_end,
		uint64_t offset, const char *name)
{
	int i;
	for (i = 0; i < mappings->elems; i++) {
		if (mappings->tab[i].m_start == mapping_start) {
			mappings->tab[i].m_end = mapping_end;
			mappings->tab[i].m_name = name;
			mappings->tab[i].m_offset = offset;
			mappings->tab[i].m_fd = open(name, O_RDONLY);
			mappings->tab[i].m_elf = elf_begin(mappings->tab[i].m_fd, ELF_C_READ_MMAP, NULL);
			return;
		}
	}
}
#endif

/**
 * @brief Gets registers from core dump
 *
 * @param core ELF handler for the core dump file
 * @param core_file_name name of the core file; needed only for diagnostics
 * @param mappings mappings database
 * @return notes handler, NULL on error
 */
static Elf_Data *__get_registers_core(Elf *core, const char *core_file_name, Mappings *mappings)
{
	Elf_Data *notes = NULL;
	/* The part below uses libebl. In case WITH_CORE_DUMP is enabled, the license
	 * may switch to GPL2.
	 */
#ifdef WITH_CORE_DUMP
	GElf_Phdr mem;
	GElf_Phdr *phdr = gelf_getphdr(core, 0, &mem);

	if (phdr == NULL || phdr->p_type != PT_NOTE) {
		fprintf(errfile, "%s : Missing note section at the first position in core file\n",
				core_file_name);
		return NULL;
	}

	notes = elf_getdata_rawchunk(core, phdr->p_offset, phdr->p_filesz, ELF_T_NHDR);
	if (notes == NULL) {
		fprintf(errfile, "%s : error getting notes (%s)\n", core_file_name, dwfl_errmsg(-1));
		return NULL;
	}

	Ebl *ebl = ebl_openbackend(core);
	if (ebl == NULL) {
		fprintf(errfile, "%s : Can't initialize ebl\n", core_file_name);
		return NULL;
	}

	GElf_Nhdr nhdr;
	size_t name_pos;
	size_t desc_pos;
	size_t pos = 0;
	size_t new_pos = 0;
	int got_regs = 0;
	/* registers should be in the first note! */
	while ((new_pos = gelf_getnote(notes, pos, &nhdr, &name_pos, &desc_pos)) > 0) {
		if (nhdr.n_type == NT_PRSTATUS && !got_regs) {
			GElf_Word regs_offset;
			size_t nregloc;
			const Ebl_Register_Location *reglocs;
			size_t nitems;
			const Ebl_Core_Item *items;

			got_regs = 1;

			if (0 == ebl_core_note(ebl, &nhdr, "CORE", &regs_offset, &nregloc,
						&reglocs, &nitems, &items)) {
				fprintf(errfile,
						"%s : error parsing notes (built with different build of libebl?)\n",
						core_file_name);
				return NULL;
			}

			const char *regs_location = (const char *)(notes->d_buf) + pos + desc_pos
				+ regs_offset;
			unsigned i;

			for (i = 0; i < nregloc; i++) {
				const char *register_location = regs_location + reglocs[i].offset;
				int regnum;
				for (regnum = reglocs[i].regno;
						regnum < reglocs[i].regno + reglocs[i].count;
						regnum++) {
					char regname[5];
					int bits, type;
					const char *prefix = 0;
					const char *setname = 0;

					ssize_t ret = ebl_register_info(ebl, regnum, regname,
							sizeof(regname), &prefix, &setname,
							&bits, &type);
					if (ret < 0) {
						fprintf(errfile, "%s : can't get register info\n", core_file_name);
						return NULL;
					}
					void *place_for_reg_value = _get_place_for_register_value(regname, regnum);

					if (place_for_reg_value != NULL)
						__get_value(core, register_location, bits, place_for_reg_value);

					register_location += bits / 8 + reglocs[i].pad;
				}
			}
		} else if (nhdr.n_type == NT_FILE) {
			uint64_t values_cnt = 0, page_size = 0;
			const char *values;
			const char *filenames;
			size_t addr_size = 0;

			__parse_note_file(core, notes->d_buf + desc_pos, &values_cnt, &page_size,
					&addr_size, &values, &filenames);

			int ii;
			/* First: triplets of <mapping-start> <mapping-end> <offset-in-pages>
			 *     count = values_cnt
			 * Then the names of files.
			 */
			for (ii = 0; ii < values_cnt; ii++) {
				uint64_t mapping_start = 0, mapping_end = 0, offset_in_pages = 0;
				const char *item = values + 3 * addr_size * ii;

				__get_mapping_item(core, addr_size, item, &mapping_start, &mapping_end,
						&offset_in_pages);
				__updateMapping(mappings, mapping_start, mapping_end,
						offset_in_pages*page_size, filenames);
				filenames += strlen(filenames)+1;
			}
		}
		pos = new_pos;
	}
	ebl_closebackend(ebl);
#else
	fprintf(errfile, "Configured without support for core dump files\n");
#endif
	return notes;
}

/**
 * @brief Prints call stack to the screen.
 *
 * @param callstack gathered call stack database
 * @param dwfl dwfl handler
 * @param core ELF handler for the core dump file, NULL if live process analyzed
 * @param pid PID of the live process, 0 if core dump file analyzed
 * @param notes notes handler, NULL if live process analyzed
 */
static void __print_callstack(Callstack *callstack, Dwfl *dwfl, Elf *core, pid_t pid,
		Elf_Data *notes)
{
	fprintf(outputfile, "Call stack");
	if (pid > 1) fprintf(outputfile, " for PID %d", pid);
	fprintf(outputfile, ":\n");

	char *dem_buffer = NULL;
	size_t it;
	for (it = 0; it != callstack->elems; ++it) {
		if (sizeof(callstack->tab[0]) > 4)
			fprintf(outputfile, "0x%016llx: ", (long long)callstack->tab[it]);
		else
			fprintf(outputfile, "0x%08x: ", (int32_t)callstack->tab[it]);
		const char *symbol = callstack->proc_name[it];

		Dwfl_Module *module = dwfl_addrmodule(dwfl, callstack->tab[it]);
		if (module) {
			char *demangled_symbol = 0;
			const char *fname = 0;
			const char *module_name = dwfl_module_info(module, NULL, NULL, NULL, NULL, NULL, &fname, NULL);
			char *symbol_from_elf = 0;
			if (symbol == NULL)
				symbol = dwfl_module_addrname(module, callstack->tab[it]);
			if (symbol == NULL)
				symbol = symbol_from_elf = __try_symbol_from_elfs(core, notes, callstack->tab[it], &fname);
			if (symbol != 0 && symbol[0] == '_' && symbol[1] == 'Z') {
				int status = -1;

				demangled_symbol = __cxa_demangle(symbol, dem_buffer, NULL, &status);
				if (status == 0)
					symbol = demangled_symbol;
			}
			if (symbol != 0)
				fprintf(outputfile, "%s()", symbol);
			else
				fprintf(outputfile, "<unknown>");

			if (demangled_symbol != 0)
				free(demangled_symbol);

			if (symbol_from_elf != 0)
				free(symbol_from_elf);

			fprintf(outputfile, " from %s\n", fname != NULL ? fname : module_name);
		} else if (symbol) {
			fprintf(outputfile, "%s()", symbol);
		} else {
			fprintf(outputfile, "unknown function\n");
		}
	}
}

/**
 * @brief Main function.
 *
 * Main module accepts two forms of launching:
 *
 *     crash-stack core-dump-file
 *     crash-stack --pid pid
 *
 * The first form allows user to print call stack of a generated core dump file.
 * The second form allows connecting to a live process and displaying its call stack.
 * It might be also used for connecting to a process from system's core dump handler.
 */
int main(int argc, char **argv)
{
	int c, i;
	pid_t pid = 0;

	const char *core_file_name;

	prctl(PR_SET_DUMPABLE, 0);

	while ((c = getopt_long_only(argc, argv, "", opts, NULL)) != -1) {
		switch (c) {
		case OPT_PID:
			pid = atoi(optarg);
			break;
		case OPT_OUTPUTFILE:
			outputfile = fopen(optarg, "w");
			break;
		case OPT_ERRFILE:
			errfile = fopen(optarg, "w");
			break;
		}
	}

	if (NULL == errfile) errfile = stderr;
	if (NULL == outputfile) outputfile = stdout;

	core_file_name = argv[optind];
	argc -= optind;

	elf_version(EV_CURRENT);

	/* First, prepare dwfl and modules */
	Elf *core = NULL;
	int core_fd = -1;
	Dwfl *dwfl = NULL;

	if (pid > 1)
		dwfl = __open_dwfl_with_pid(pid);
	else {
		if (argc != 1) {
			fprintf(errfile,
					"Usage: %s [--output file] [--erroutput file] [--pid <pid> | <core-file>]\n",
					argv[0]);
			return 1;
		}

		core_fd = open(core_file_name, O_RDONLY);
		if (core_fd < 0) {
			perror(core_file_name);
			return 2;
		}

		core = elf_begin(core_fd, ELF_C_READ_MMAP, NULL);
		if (core == NULL) {
			fprintf(errfile, "%s : Can't open ELF (%s)\n", core_file_name, elf_errmsg(-1));
			return 3;
		}

		dwfl = __open_dwfl_with_core(core, core_file_name);
	}

	if (NULL == dwfl)
		return 1111;

	Mappings mappings;
	mappings.elems = 0;

	dwfl_getmodules(dwfl, __module_callback, &mappings, 0);
	Elf_Data *notes = 0;

	/* Now, get registers */
	if (pid > 1) {
		if (-1 == __get_signal_ptrace(pid))
			return 4444;
		if (-1 == __get_registers_ptrace(pid))
			return 3333;
	} else {
		notes = __get_registers_core(core, core_file_name, &mappings);
		if (NULL == notes)
			return 2222;
	}

	/* Unwind call stack */
	Callstack callstack;
	callstack.elems = 0;
	memset(callstack.proc_name, 0, sizeof(callstack.proc_name));

	_create_crash_stack(dwfl, core, pid, &mappings, &callstack);

	/* Print registers */
	_crash_stack_print_regs(outputfile);

	/* Print the results */
	__print_callstack(&callstack, dwfl, core, pid, notes);

	for (i = 0; i < callstack.elems; ++i)
		if (callstack.proc_name[i])
			free(callstack.proc_name[i]);

	/* Clean up */
	dwfl_report_end(dwfl, NULL, NULL);
	dwfl_end(dwfl);
	if (NULL != core) elf_end(core);
	if (-1 != core_fd) close(core_fd);

	return 0;
}
